<?php
/**
 * CGI5.php - A CGI-request parameter management class for PHP 5+
 *
 * This module was originally based on Perl's CGI module. Only
 * parameter-related functionality is included.
 *
 * For more information, contact:
 *
 *	Kappa Computer Solutions, LLC
 *	http://www.kappacs.com/
 *
 * 2011-10-24	PHP 5-only version with full request tree support
 *		(derived from Brian Katzung's CGI.php) by Brian Katzung
 *
 * @copyright Copyright 2011 Brian Katzung and Kappa Computer Solutions, LLC
 * @author Brian Katzung <briank@kappacs.com>
 * @license http://www.opensource.org/licenses/MIT MIT License
 * @package CGI
 */

require_once('ArrayPath.php');

/**
 * Manage CGI POST and GET request parameters
 */
class CGI5 implements ArrayAccess, Iterator
{
    /**
     * Instance variables
     */

    /**
     * @var &array The root of the request tree.
     */
    protected $request;

    /**
     * @var ArrayPath The default viewpoint for parameter names.
     */
    protected $viewpoint;

    /**
     * Class methods
     */

    /**
     * Constructor
     *
     * @param array &$request An optional reference to a source of request
     *  paramenters (the default is $_REQUEST).
     */
    function __construct (&$request = null) {
	if (is_array($request)) {
	    /* Use the specified request tree */
	    $this->request = &$request;
	} else {
	    /* Default to $_REQUEST */
	    $this->request = &$_REQUEST;
	}
	$this->viewpoint = null;
    }

    /**
     * Non-HTML-generating instance methods
     */

    /**
     * Append values to a parameter.
     *
     * @param mixed $name The parameter to which to append the values.
     * @param mixed $what The value(s) to append.
     * @param mixed $viewpoint An optional alternative viewpoint.
     * @return CGI5
     */
    function append ($name, $what, $viewpoint = null) {
	/* Promote $what to array if scalar */
	$awhat = (array) $what;

	/* Quit now if there's nothing to append */
	if (!count($awhat)) { return $this; }

	$name = ArrayPath::get($name, $this->getViewpoint($viewpoint));

	/* Find (or create) the branch onto which to append */
	$branch = &$name->apply($this->request,
	  array('create' => true, 'partial' => true, 'last' => &$last));

	if (isset($last) && '' !== $last) {
	    if (!isset($branch[$last])) {
		/* No existing value; just use the appended value(s) */
		$branch[$last] = $what;
		return $this;
	    }

	    if (!is_array($branch[$last])) {
		/*
		 * Promote the existing scalar to an array
		 * in preparation for merging.
		 */
		$branch[$last] = array($branch[$last]);
	    }
	    $branch = &$branch[$last];
	}

	/* Combine the existing and new values */
	$branch = array_merge($branch, $awhat);

	return $this;
    }

    /**
     * Delete specific request parameters.
     *
     * @param mixed $what Parameter name(s) or path(s) to be deleted.
     * @param mixed $viewpoint An optional alternative viewpoint.
     * @param boolean $multiple If true, the first level of array
     *  $what constitutes multiple parameters to be deleted, not path
     *  components. The default is false.
     * @return CGI5
     */
    function delete ($what, $viewpoint = null, $multiple = false) {
	if ($multiple) {
	    foreach ($what as $name) {
		$this->delete($name, $viewpoint);
	    }
	} else {
	    $name = ArrayPath::get($name, $this->getViewpoint($viewpoint));

	    $branch = &$name->apply($this->request,
	      array('partial' => true, 'last' => &$last));

	    if (is_array($branch) && isset($last) && '' !== $last) {
		unset($branch[$last]);
	    }
	}

	return $this;
    }

    /**
     * Delete all request parameters at (or below) the current viewpoint.
     *
     * @return CGI5
     */
    function delete_all () {
      return $this->delete($this->getKeys(), null, true);
    }
    /**
     * @return CGI5
     */
    function deleteAll () {
      return $this->delete($this->getKeys(), null, true);
    }

    /**
     * Returns a *reference* to a parameter value.
     *
     * @param mixed $name The parameter name or path to fetch (the
     *  default is the entire request array).
     * @param mixed $viewpoint An optional alternate viewpoint.
     * @return mixed
     */
    public function &get ($name = '', $viewpoint = null) {
	$name = ArrayPath::get($name, $this->getViewpoint($viewpoint));
	return $name->apply($this->request);
    }

    /**
     * Shortcut for a list of array (parameter) keys.
     *
     * @param mixed $name The name of the parameter to fetch (the default
     *  is the entire request array).
     * @param mixed $viewpoint An optional alternate viewpoint.
     * @return array
     */
    public function getKeys ($name = '', $viewpoint = null) {
	$value = $this->get($name, $viewpoint);
	return is_array($value)? array_keys($value): array();
    }

    /**
     * Return parameter names or a parameter value, or set a parameter
     * value.
     *
     * @param mixed $name The optional name or path of the parameter to
     *  get or set.
     * @param mixed $value The value to set, if supplied.
     * @return mixed
     */
    public function param ($name = null, $value = null, $viewpoint = null) {
	$viewpoint = $this->getViewpoint($viewpoint);

	/* Param without a name is equivalent to getKeys. */
	if (!isset($name)) { return $this->getKeys('', $viewpoint); }

	$name = ArrayPath::get($name, $viewpoint);

	/* Param with a value is a set; without is a get. */
	if (isset($value)) { return $this->set($name, $value, $viewpoint); }
	return $this->get($name, $viewpoint);
    }

    /**
     * Set a parameter value.
     *
     * @param object|array|string $name The parameter name or path.
     * @param mixed $value The new parameter value.
     * @param mixed $viewpoint An optional alternate viewpoint.
     * @return CGI5
     */
    public function set ($name, $value, $viewpoint = null) {
	if (!isset($name)) { $name = '[]'; }
	$name = ArrayPath::get($name,  $this->getViewpoint($viewpoint));
	if ($name->isAppend()) {
	    if (is_array($branch = &$name->apply($this->request,
	      array('create' => true, 'needArray' => true)))) {
		$branch[] = $value;
	    }
	} else {
	    if (is_array($branch = &$name->apply($this->request,
	      array('create' => true, 'needArray' => true,
	      'partial' => true, 'last' => &$last)))) {
		$branch[$last] = $value;
	    }
	}
	return $this;
    }

    /**
     * Get the current viewpoint or an alternate.
     *
     * @param object|array|string $viewpoint An optional alternate viewpoint.
     * @return object
     */
    public function getViewpoint ($viewpoint = null) {
	if (!isset($viewpoint)) {
	    /*
	     * No alternate was supplied. Allocate a default viewpoint
	     * if necessary and use it.
	     */
	    if (!$this->viewpoint) { $this->viewpoint = new ArrayPath(''); }
	    return $this->viewpoint;
	}

	return ArrayPath::get($viewpoint);
    }

    /**
     * Set the current viewpoint.
     *
     * @param mixed $viewpoint The new viewpoint.
     * @return CGI5
     */
    public function setViewpoint ($viewpoint) {
	if (!isset($viewpoint)) { $this->viewpoint = null; }
	else { $this->viewpoint = ArrayPath::get($viewpoint); }

	return $this;
    }

    /**
     * HTML-generating instance methods
     */

    /**
     * Return HTML for a checkbox input
     *
     * Tag-specific attributes: <br>
     *	"checked": if the checkbox should be checked <br>
     *	"unchecked": the value when *not* checked
     *
     * Other attributes (except "override") are passed through to the
     * HTML element.
     *
     * @param array $attr Array of attributes for the input
     * @return string
     */
    public function checkbox ($attr) {
	/**
	 * Three types of checkbox symantics are supported:
	 * bi-value and traditional uni-value specifically-named
	 * checkboxes, and uni-value append-named ("[]") checkboxes.
	 *
	 * A bi-value checkbox has a hidden unchecked value.
	 *
	 * Uni-value specifically-named checkboxes are tracked in a
	 * _checkbox list so that we know when to mark them checked or
	 * unchecked.
	 */

	$hidden = '';
	if (isset($attr['name'])) {
	    $name = $attr['name'] = ArrayPath::get($attr['name'],
	      $this->getViewpoint());

	    if (!$name->isAppend()) {
		/*
		 * This is a specific name rather than an append name.
		 * Get the current value, if any.
		 */
		$value = $this->get($name);

		/* Get the list of uni-valued checkboxes */
		$checkName = new ArrayPath('_checkbox[]', true);
		$boxes = (array) $this->get($checkName);

		/* Try to use the existing setting when not overriding */
		if (!$attr['override'] && (isset($value) ||
		  (false !== array_search($name->toString(false), $boxes)))) {
		    if ($value == $attr['value']) {
			$attr['checked'] = 'checked';
		    } else {
			unset($attr['checked']);
		    }
		}

		if (isset($attr['unchecked'])) {
		    # Bi-value checkbox
		    $hidden = $this->hidden(array('name' => $name,
		      'value' => $attr['unchecked'], 'override' => 1));
		} else {
		    # Uni-value checkbox
		    $hidden = $this->hidden(array('name' => $checkName,
		      'value' => $name->toString(false), 'override' => 1));
		}
	    }
	}

	unset($attr['override'], $attr['unchecked']);
	return $hidden .
	  $this->makeTag('input type="checkbox"', $attr, ' /');

    }

    /**
     * Return HTML for a hidden input
     *
     * @param array $attr Array of attributes for the input
     * @param mixed $viewpoint An alternative viewpoint
     * @return string
     */
    public function hidden ($attr, $viewpoint = null) {
	if (isset($attr['name'])) {
	    $name = ArrayPath::get($attr['name'],
	      $this->getViewpoint($viewpoint));

	    /* Get the current value or assigned value */
	    $current = $this->get($name);
	    if (!$attr['override'] && isset($current)) { $value = $current; }
	    else { $value = $attr['value']; }

	    /* Handle the no-value case */
	    if (!isset($value)) { return ''; }

	    /* Handle the array (or tree) case */
	    if (is_array($value)) {
		$parts = array();
		foreach ($value as $subname => $subvalue) {
		    $attr['name'] = array($subname);
		    $attr['value'] = &$subvalue;
		    $parts[] = $this->hidden($attr, $name);
		}
		return join('', $parts);
	    }

	    $attr['name'] = $name;
	}

	return $this->makeValueTag('input type="hidden"', $attr, ' /');
    }

    /**
     * Helper function to return HTML for a tag with optional attributes
     *
     * @param string $html The HTML tag (element) text
     * @param array $attr Array of attributes for the tag
     * @param string $close The closing text ('' or ' /') for the tag
     * @return string
     */
    protected function makeTag ($html, $attr, $close = '') {
	if (isset($attr['name'])) {
	    /* Expand the name to the full path */
	    $attr['name'] = ArrayPath::get($attr['name'],
	      $this->getViewpoint())->toString(true);
	}

	foreach ($attr as $key => $val) {
	    if ('' !== $key && false !== $val)
	    {
		$html .= (null === $val || true === $val)? " $key":
		  " $key=\"" . $this->quote($val) . "\"";
	    }
	}

	return "<$html$close>";
    }

    /**
     * Helper function to return HTML for a tag using the request value
     *
     * @param string $html The HTML tag (element) text
     * @param array $attr Array of attributes for the tag
     * @param string $close The closing text ('' or ' /') for the tag
     * @return string
     */
    protected function makeValueTag ($html, $attr, $close = '') {
	if (!$attr['override'] && isset($attr['name'])) {
	    $value = $this->get($attr['name']);
	    if (isset($value)) { $attr['value'] = $value; }
	}
	unset($attr['override']);

	return $this->makeTag($html, $attr, $close);
    }

    /**
     * Return HTML for a password input
     *
     * @param array $attr Array of attributes for the tag
     * @return string
     */
    function password ($attr) {
	return $this->makeValueTag('input type="password"', $attr, ' /');
    }

    /**
     * Return HTML for a pop-up menu (a selection of size one)
     *
     * @param array $attr Array of attributes for the tag
     * @return string
     */
    function popup_menu ($attr) {
	$options = $attr['options'];
	if (isset($options)) {
	    $labels = array_keys($options);
	    $values = array_values($options);
	} else {
	    $values = $attr['values'];
	    $labels = $attr['labels'];
	}
	$default = (array) $attr['default'];
	unset($attr['options'], $attr['labels'], $attr['values'],
	  $attr['default']);

	if (isset($attr['name'])) {
	    $name = $attr['name'] = ArrayPath::get($attr['name'],
	      $this->getViewpoint());

	    $current = $this->get($name);
	    if (!$attr['override'] && isset($current)) {
		$default = (array) $current;
	    }

	    if (array_key_exists('multiple', $attr)) {
		/* Make sure the parameter will support multiple values */
		$name->isAppend(true);
	    }
	}
	unset($attr['override']);

	$html = array();
	$html[] = $this->makeTag('select', $attr);

	foreach ($values as $i => $value) {
	    $sel = (false !== array_search($value, $default))?
	      ' selected="selected"': '';
	    if (isset($labels[$i])) {
		$html[] = '<option value="' .
		  $this->quote($value) .  "\"$sel>" .
		  $this->quote($labels[$i]) . '</option>';
	    } else {
		$html[] = "<option$sel>" .
		  $this->quote($value) . "</option>";
	    }
	}

	$html[] = '</select>';
	return join('', $html);
    }

    /**
     * @param array $attr
     * @return string
     */
    function popupMenu ($attr) { return $this->popup_menu($attr); }

    /**
     * Quote a text string
     *
     * @param string $value The text to quote.
     * @return string
     */
    function quote ($value) {
	return htmlspecialchars($value, ENT_QUOTES);
    }

    /**
     * Return HTML for a (single) radio button
     *
     * @param array $attr Array of attributes for the tag.
     * @return string
     */
    function radiobutton ($attr) {
	$name = $attr['name'];
	$value = $this->get($name);
	if (!$attr['override'] && isset($value)) {
	    if ($value == $attr['value']) {
		$attr['checked'] = 'checked';
	    } else {
		unset($attr['checked']);
	    }
	}
	unset($attr['override']);

	return $this->makeTag('input type="radio"', $attr, ' /');
    }

    /**
     * Return HTML for a scrolling list (a selection of size greater than one)
     *
     * @param array $attr Array of attributes for the tag
     * @return string
     */
    function scrolling_list ($attr) {
	if (!isset($attr['size'])) {
	    $attr['size'] = 2;
	}

	return $this->popup_menu($attr);
    }

    /**
     * @param array $attr
     * @return string
     */
    function scrollingList ($attr) { return $this->scrolling_list($attr); }

    /**
     * Return HTML for a text area
     *
     * @param array $attr Array of attributes for the tag
     * @return string
     */
    function textarea ($attr) {
	$name = $attr['name'];
	$current = $this->get($name);
	if (!$attr['override'] && isset($current)) { $value = $current; }
	else { $value = $attr['value']; }
	unset($attr['value'], $attr['override']);

	return $this->makeTag('textarea', $attr) . $this->quote($value) .
	  '</textarea>';
    }

    /**
     * Return HTML for a text field
     *
     * @param array $attr Array of attributes for the tag
     * @return string
     */
    function textfield ($attr) {
	return $this->makeValueTag('input type="text"', $attr, ' /');
    }

    /**
     * Magic member interface methods
     */

    /**
     * Returns a *reference* to a request parameter as an object member.
     *
     * @param string $name The name of the parameter
     * @return mixed
     */
    public function &__get ($name) { return $this->get($name); }

    /**
     * Set a request parameter as an object member.
     *
     * @param string $name The name of the parameter
     * @param mixed $value The value of the parameter
     * @return mixed
     */
    public function __set ($name, $value) { $this->set($name, $value); }

    /**
     * Return whether the parameter is set.
     *
     * @param string $name
     * @return boolean
     */
    public function __isset ($name) { return $this->offsetExists($name); }

    /**
     * Delete the parameter.
     *
     * @param string $name
     */
    public function __unset ($name) { $this->delete(array($name)); }

    /**
     * ArrayAccess interface methods
     */

    /**
     * Return a request parameter as an array element.
     *
     * @param string $name The name of the parameter
     * @return mixed
     */
    public function offsetGet ($name) { return $this->get($name); }

    /**
     * Set a request parameter as an array element.
     *
     * @param string|null $name The name of the parameter
     * @param mixed $value The value of the parameter
     */
    public function offsetSet ($name, $value) { $this->set($name, $value); }

    /**
     * Return whether the parameter is set.
     *
     * @param string $name
     * @return boolean
     */
    public function offsetExists ($name) {
	$value = $this->get($name);
	return isset($value);
    }

    /**
     * Delete the parameter.
     *
     * @param string $name
     */
    public function offsetUnset ($name) { $this->delete($name); }

    /**
     * Iterator interface methods
     */

    /**
     * Rewind the iterator.
     */
    public function rewind () {
	reset($this->getViewpoint()->apply($this->request));
    }

    /**
     * Return the current key.
     *
     * @return string|null
     */
    public function key () {
	return key($this->getViewpoint()->apply($this->request));
    }

    /**
     * Return whether the current position is valid.
     *
     * @return boolean
     */
    public function valid () { return null !== $this->key(); }

    /**
     * Return the current parameter or cover object.
     *
     * @return mixed
     */
    public function current () {
	return current($this->getViewpoint()->apply($this->request));
    }

    /**
     * Return the next parameter or cover object.
     *
     * @return mixed
     */
    public function next () {
	return next($this->getViewpoint()->apply($this->request));
    }

}

# END
